import os
import shutil
import ssl
import subprocess
import urllib
import yaml

from base64 import b64decode

from charms.reactive import (
    set_state,
    remove_state,
    when,
    when_not,
    when_any,
)

from charmhelpers.core import (
    hookenv,
    host,
)

from charms.layer import snap

from charms.leadership import (
    leader_get,
    leader_set,
)

from pymongo import MongoClient
from pymongo.errors import (
    DuplicateKeyError,
    NotMasterError,
    )

from tempfile import NamedTemporaryFile

SNAP_DATA = '/var/snap/prometheus-mongodb-exporter/current/'
CA_PATH = os.path.join(SNAP_DATA, 'cacert.pem')
CERT_PATH = os.path.join(SNAP_DATA, 'server.pem')

ARGS_PATH = '/var/snap/prometheus-mongodb-exporter/current/daemon_arguments'
SYSTEMD_SERVICE = 'snap.prometheus-mongodb-exporter.mongodb-exporter.service'

JUJU_DB_URI = 'juju-db://'

JUJU_ROOT = '/var/lib/juju'

MIN_PORT = 1
MAX_PORT = 65535
UNSENSIBLE_PORT_MESSAGE = 'port must be between {} and {}'.format(MIN_PORT, MAX_PORT)


def blocked(message):
        hookenv.log(message)
        hookenv.status_set('blocked', message)


def mongodb_uri_is_juju_db():
    return hookenv.config()['mongodb_uri'] == JUJU_DB_URI


def get_machine_agent_conf():
    agent_conf_path = os.path.join(
        JUJU_ROOT, 'agents', 'machine-{}'.format(os.environ['JUJU_MACHINE_ID']), 'agent.conf')
    with open(agent_conf_path) as f:
        return yaml.safe_load(f.read())


def get_mongodb_uri():
    if mongodb_uri_is_juju_db():
        return leader_get('mongodb_uri')
    return hookenv.config()['mongodb_uri']


@when('config.changed.mongodb_uri')
def maybe_monitor_juju_db():
    if mongodb_uri_is_juju_db():
        set_state('prometheus-mongodb-exporter.monitor-juju-db')
    else:
        remove_state('prometheus-mongodb-exporter.monitor-juju-db')


@when('prometheus-mongodb-exporter.monitor-juju-db')
@when('leadership.is_leader')
@when_not('leadership.set.mongodb_uri')
def create_juju_db_user():
    agent_conf = get_machine_agent_conf()

    with NamedTemporaryFile() as cacert:
        cacert.write(agent_conf['cacert'].encode('utf-8'))
        cacert.flush()

        dbusername = agent_conf['tag']
        dbpassword = urllib.parse.quote(agent_conf['statepassword'], safe='')
        db = MongoClient(
            'mongodb://{}:{}@localhost:37017/admin?replicaSet=juju'.format(dbusername, dbpassword),
            ssl_ca_certs=cacert.name,
            ssl_cert_reqs=ssl.CERT_REQUIRED,
            ssl_match_hostname=False,  # Juju's server.pem no longer includes member IPs
            )

        username = 'prometheus-mongodb-exporter'
        password = host.pwgen()
        roles = [
            {"role": "clusterMonitor", "db": "admin"},
            {"role": "read", "db": "local"},
            ]
        try:
            db.admin.command("createUser", username, pwd=password, roles=roles)
        except DuplicateKeyError:
            # The user already exists, but mongodb_uri is not set?
            # Well, then, let's just reset the password and roles.
            hookenv.log('create_juju_db_user: user {} already exists; resetting password and roles'.format(username))
            db.admin.command('updateUser', username, pwd=password, roles=roles)
        except NotMasterError:
            blocked('mongodb master unavailable - user not created')
            # Exit quietly w/o a fuss; skipping leader_set will cause a retry.
            return

    leader_set(mongodb_uri='mongodb://{}:{}@localhost:37017/admin'.format(username, password))


@when_not('prometheus-mongodb-exporter.installed')
def install_prometheus_mongodb_exporter():
    ''' Install the proxy application '''
    ca = hookenv.config()['mongodb_tls_ca']
    cert = hookenv.config()['mongodb_tls_cert']
    uri = get_mongodb_uri()

    snap.install('prometheus-mongodb-exporter')

    if not uri:
        blocked('mongodb_uri must be set')
        return

    # Create a systemd unit file, and open the port.
    port = hookenv.config()['port']
    if not sensible_port(port):
        blocked(UNSENSIBLE_PORT_MESSAGE)
        return

    write_daemon_args(port, uri, ca, cert)
    write_certificates(ca, cert)
    hookenv.open_port(port)
    hookenv.status_set('active', 'serving metrics')
    set_state('prometheus-mongodb-exporter.installed')


def sensible_port(port):
    return (port < MAX_PORT and port > MIN_PORT)


@when('config.changed.port')
def update_opened_port():
    config = hookenv.config()

    new_port = config['port']
    old_port = config.previous('port')

    if old_port and sensible_port(old_port):
        hookenv.close_port(old_port)

    if not sensible_port(new_port):
        blocked(UNSENSIBLE_PORT_MESSAGE)
        return

    hookenv.open_port(new_port)


@when('prometheus-mongodb-exporter.installed')
@when_any('config.changed', 'leadership.changed')
def update_daemon_args():
    config = hookenv.config()
    port = config['port']
    uri = get_mongodb_uri()
    ca = config['mongodb_tls_ca']
    cert = config['mongodb_tls_cert']

    if not sensible_port(port):
        blocked(UNSENSIBLE_PORT_MESSAGE)
        return

    write_daemon_args(port, uri, ca, cert)
    subprocess.check_call(['systemctl', 'daemon-reload'])
    set_state('prometheus-mongodb-exporter.do-restart')


def write_daemon_args(port, mongodb_uri, ca, cert):
    listen_addr = ':{}'.format(port)
    args_dict = {'web.listen-address': listen_addr,
                 'mongodb.uri': mongodb_uri}

    if mongodb_uri_is_juju_db():
        args_dict['mongodb.tls-ca'] = CA_PATH
        args_dict['mongodb.tls-cert'] = CERT_PATH

    # Handle the CA
    if ca:
        args_dict['mongodb.tls-ca'] = CA_PATH

    # Handle the certificate
    if cert:
        args_dict['mongodb.tls-cert'] = CERT_PATH

    args = ''
    for k, v in args_dict.items():
        args = '{} -{} {}'.format(args, k, v)

    args = 'ARGS="{}"'.format(args)

    with open(ARGS_PATH, 'w') as f:
        f.write(args)

    set_state('prometheus-mongodb-exporter.do-restart')


def write_certificates(ca, cert):
    if mongodb_uri_is_juju_db():
        write_ca('juju')
        write_client_cert('juju')
    else:
        write_ca(ca)
        write_client_cert(cert)


def write_ca(ca):
    if ca:
        if ca == 'juju':
            agent_conf = get_machine_agent_conf()
            ca_cleartext = agent_conf['cacert'].encode('UTF-8')
        else:
            ca_cleartext = b64decode(ca)

        host.write_file(CA_PATH, ca_cleartext, perms=0o400)
    else:
        try:
            os.remove(CA_PATH)
        except FileNotFoundError:
            pass


def write_client_cert(cert):
    if cert:
        if cert == 'juju':
            shutil.copy(os.path.join(JUJU_ROOT, 'server.pem'), CERT_PATH)
        else:
            host.write_file(CERT_PATH, b64decode(cert), perms=0o400)
    else:
        try:
            os.remove(CERT_PATH)
        except FileNotFoundError:
            pass


@when('prometheus-mongodb-exporter.installed')
@when_any('config.changed.mongodb_tls_ca', 'config.changed.mongodb_tls_cert')
def update_certificates():
    write_certificates(hookenv.config()['mongodb_tls_ca'],
                       hookenv.config()['mongodb_tls_cert'])

    set_state('prometheus-mongodb-exporter.do-restart')


@when('prometheus-mongodb-exporter.do-restart')
def do_restart():
    hookenv.log('Restarting prometheus-mongodb-exporter')
    hookenv.status_set('active', 'serving metrics')
    subprocess.check_call(['systemctl', 'restart',
                           SYSTEMD_SERVICE])
    remove_state('prometheus-mongodb-exporter.do-restart')
